/*
 * Written by Solar Designer and placed in the public domain.
 * See crypt_blowfish.c for more information.
 */

#include <string.h>

#ifdef TEST
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#endif

#define CRYPT_OUTPUT_SIZE		(7 + 22 + 31 + 1)
#define CRYPT_GENSALT_OUTPUT_SIZE	(7 + 22 + 1)

#if defined(__GNU_LIBRARY__) && defined(_LIBC)
#define __SKIP_GNU
#endif
#include "ow-crypt.h"

extern char *_crypt_blowfish_rn(__CONST char *key, __CONST char *salt,
	char *output, int size);
extern char *_crypt_gensalt_blowfish_rn(unsigned long count,
	__CONST char *input, int isize, char *output, int osize);

extern unsigned char _crypt_itoa64[0x40];
extern char *_crypt_gensalt_traditional_rn(unsigned long count,
	__CONST char *input, int isize, char *output, int osize);
extern char *_crypt_gensalt_extended_rn(unsigned long count,
	__CONST char *input, int isize, char *output, int osize);
extern char *_crypt_gensalt_md5_rn(unsigned long count,
	__CONST char *input, int isize, char *output, int osize);

#if defined(__GNU_LIBRARY__) && defined(_LIBC)
/* crypt.h from glibc-crypt-2.1 will define struct crypt_data for us */
#include "crypt.h"
extern char *__md5_crypt_r(const char *key, const char *salt,
	char *buffer, int buflen);
/* crypt-entry.c needs to be patched to define __des_crypt_r rather than
 * __crypt_r, and not define crypt_r and crypt at all */
extern char *__des_crypt_r(const char *key, const char *salt,
	struct crypt_data *data);
extern struct crypt_data _ufc_foobar;
#endif

#if defined(__GNU_LIBRARY__) && defined(_LIBC)
char *__crypt_rn(__const char *key, __const char *salt, void *data, int size)
{
	if (salt[0] == '$' && salt[1] == '2')
		return _crypt_blowfish_rn(key, salt, (char *)data, size);
	if (salt[0] == '$' && salt[1] == '1')
		return __md5_crypt_r(key, salt, (char *)data, size);
	if (size >= sizeof(struct crypt_data))
		return __des_crypt_r(key, salt, (struct crypt_data *)data);
	return NULL;
}

char *__crypt_r(__const char *key, __const char *salt, struct crypt_data *data)
{
	return __crypt_rn(key, salt, data, sizeof(*data));
}

char *__crypt(__const char *key, __const char *salt)
{
	return __crypt_rn(key, salt, &_ufc_foobar, sizeof(_ufc_foobar));
}
#else
char *crypt_rn(__CONST char *key, __CONST char *salt, void *data, int size)
{
	return _crypt_blowfish_rn(key, salt, (char *)data, size);
}

char *crypt_r(__CONST char *key, __CONST char *salt, void *data)
{
	return crypt_rn(key, salt, data, CRYPT_OUTPUT_SIZE);
}

char *crypt(__CONST char *key, __CONST char *salt)
{
	static char output[CRYPT_OUTPUT_SIZE];

	return crypt_rn(key, salt, output, sizeof(output));
}

#define __crypt_gensalt_rn crypt_gensalt_rn
#define __crypt_gensalt crypt_gensalt
#endif

char *__crypt_gensalt_rn(__CONST char *prefix, unsigned long count,
	__CONST char *input, int isize, char *output, int osize)
{
	char *(*use)(unsigned long count,
		__CONST char *input, int isize, char *output, int osize);

	if (!strncmp(prefix, "$2a$", 4))
		use = _crypt_gensalt_blowfish_rn;
	else
	if (!strncmp(prefix, "$1$", 3))
		use = _crypt_gensalt_md5_rn;
	else
	if (prefix[0] == '_')
		use = _crypt_gensalt_extended_rn;
	else
	if (!prefix[0] ||
	    (prefix[0] && prefix[1] &&
	    memchr(_crypt_itoa64, prefix[0], sizeof(_crypt_itoa64)) &&
	    memchr(_crypt_itoa64, prefix[1], sizeof(_crypt_itoa64))))
		use = _crypt_gensalt_traditional_rn;
	else
		return NULL;

	return use(count, input, isize, output, osize);
}

char *__crypt_gensalt(__CONST char *prefix, unsigned long count,
	__CONST char *input, int size)
{
	static char output[CRYPT_GENSALT_OUTPUT_SIZE];

	return __crypt_gensalt_rn(prefix, count,
		input, size, output, sizeof(output));
}

#if defined(__GNU_LIBRARY__) && defined(_LIBC)
weak_alias(__crypt_rn, crypt_rn)
weak_alias(__crypt_r, crypt_r)
weak_alias(__crypt, crypt)
weak_alias(__crypt_gensalt_rn, crypt_gensalt_rn)
weak_alias(__crypt_gensalt, crypt_gensalt)
#endif

#ifdef TEST
static struct {
	char *hash;
	char *pw;
} tests[] = {
	{"$2a$05$CCCCCCCCCCCCCCCCCCCCC.E5YPO9kmyuRGyh0XouQYb4YMJKvyOeW",
		"U*U"},
	{"$2a$05$CCCCCCCCCCCCCCCCCCCCC.VGOzA784oUp/Z0DY336zx7pLYAy0lwK",
		"U*U*"},
	{"$2a$05$XXXXXXXXXXXXXXXXXXXXXOAcXxm9kjPGEMsLznoKqmqw7tc8WCx4a",
		"U*U*U"},
	{"$2a$05$CCCCCCCCCCCCCCCCCCCCC.7uG0VCzI2bS7j6ymqJi9CdcdxiRTWNy",
		""},
	{"$2a$05$abcdefghijklmnopqrstuu5s2v8.iXieOjg/.AySBTTZIIVFJeBui",
		"0123456789abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"},
	{NULL}
};

#define which				tests[0]

static volatile int running;

static void handle_timer(int signum)
{
	running = 0;
}

int main(void)
{
	struct itimerval it;
	struct tms buf;
	clock_t start_real, start_virtual, end_real, end_virtual;
	unsigned long count;
	int i;

	for (i = 0; tests[i].hash; i++)
	if (strcmp(crypt(tests[i].pw, tests[i].hash), tests[i].hash)) {
		puts("FAILED");
		return 1;
	}

	running = 1;
	signal(SIGALRM, handle_timer);

	memset(&it, 0, sizeof(it));
	it.it_value.tv_sec = 5;
	setitimer(ITIMER_REAL, &it, NULL);

	start_real = times(&buf);
	start_virtual = buf.tms_utime + buf.tms_stime;
	count = 0;

	do {
		crypt(which.pw, which.hash);
		count++;
	} while (running);

	end_real = times(&buf);
	end_virtual = buf.tms_utime + buf.tms_stime;
	if (end_virtual == start_virtual) end_virtual++;

	printf("%.1f c/s real, %.1f c/s virtual\n",
		(float)count * CLK_TCK / (end_real - start_real),
		(float)count * CLK_TCK / (end_virtual - start_virtual));

	return 0;
}
#endif
